@@ -419,8 +419,13 @@ QWaylandInputDevice::QWaylandInputDevice(QWaylandDisplay *display, int version,
 #endif
 }
 
-// Can't be in header because dtors for scoped pointers aren't known there.
-QWaylandInputDevice::~QWaylandInputDevice() = default;
+QWaylandInputDevice::~QWaylandInputDevice()
+{
+    if (version() >= WL_SEAT_RELEASE_SINCE_VERSION)
+        release();
+    else
+        wl_seat_destroy(object());
+}
 
 void QWaylandInputDevice::seat_capabilities(uint32_t caps)
 {
@@ -523,12 +528,6 @@ void QWaylandInputDevice::handleEndDrag()
         mPointer->releaseButtons();
 }
 
-void QWaylandInputDevice::handleStartDrag()
-{
-    if (mPointer)
-        mPointer->leavePointers();
-}
-
 #if QT_CONFIG(wayland_datadevice)
 void QWaylandInputDevice::setDataDevice(QWaylandDataDevice *device)
 {
@@ -558,18 +557,6 @@ void QWaylandInputDevice::setTextInput(QWaylandTextInputInterface *textInput)
     mTextInput.reset(textInput);
 }
 
-#if QT_CONFIG(tabletevent)
-void QWaylandInputDevice::setTabletSeat(QWaylandTabletSeatV2 *tabletSeat)
-{
-    mTabletSeat.reset(tabletSeat);
-}
-
-QWaylandTabletSeatV2 *QWaylandInputDevice::tabletSeat() const
-{
-    return mTabletSeat.get();
-}
-#endif
-
 void QWaylandInputDevice::setTextInputMethod(QWaylandTextInputMethod *textInputMethod)
 {
     mTextInputMethod.reset(textInputMethod);
@@ -899,24 +886,14 @@ void QWaylandInputDevice::Pointer::releaseButtons()
     }
 }
 
-void QWaylandInputDevice::Pointer::leavePointers()
-{
-    if (auto *window = focusWindow()) {
-        LeaveEvent e(focusWindow(), mSurfacePos, mGlobalPos);
-        window->handleMouse(mParent, e);
-    }
-}
-
 class WheelEvent : public QWaylandPointerEvent
 {
 public:
     WheelEvent(QWaylandWindow *surface, Qt::ScrollPhase phase, ulong timestamp, const QPointF &local,
                const QPointF &global, const QPoint &pixelDelta, const QPoint &angleDelta,
                Qt::MouseEventSource source, Qt::KeyboardModifiers modifiers, bool inverted)
-        : QWaylandPointerEvent(QEvent::Wheel, phase, surface, timestamp, local, global,
-                               modifiers & Qt::AltModifier ? pixelDelta.transposed() : pixelDelta,
-                               modifiers & Qt::AltModifier ? angleDelta.transposed() : angleDelta,
-                               source, modifiers, inverted)
+        : QWaylandPointerEvent(QEvent::Wheel, phase, surface, timestamp,
+                               local, global, pixelDelta, angleDelta, source, modifiers, inverted)
     {
     }
 };
@@ -1583,9 +1560,9 @@ void QWaylandInputDevice::Touch::touch_frame()
         const QWindowSystemInterface::TouchPoint &tp = mPendingTouchPoints.constLast();
         // When the touch event is received, the global pos is calculated with the margins
         // in mind. Now we need to adjust again to get the correct local pos back.
-        QMargins margins = window->frameMargins();
+        QMargins margins = mFocus->clientSideMargins();
         QPoint p = tp.area.center().toPoint();
-        QPointF localPos(window->mapFromGlobal(QPoint(p.x() + margins.left(), p.y() + margins.top())));
+        QPointF localPos(mFocus->mapFromGlobal(p) + QPoint(margins.left(), margins.top()));
         if (mFocus->touchDragDecoration(mParent, localPos, tp.area.center(), tp.state, mParent->modifiers()))
             return;
     }
