@@ -6,11 +6,13 @@
 #include "qwaylandxdgshell_p.h"
 
 #include "qwaylandxdgexporterv2_p.h"
+#include "qwaylandxdgdialogv1_p.h"
 
 #include <QtWaylandClient/private/qwaylanddisplay_p.h>
 #include <QtWaylandClient/private/qwaylandwindow_p.h>
 #include <QtWaylandClient/private/qwaylandinputdevice_p.h>
 #include <QtWaylandClient/private/qwaylandscreen_p.h>
+#include <QtWaylandClient/private/qwaylandcursor_p.h>
 #include <QtWaylandClient/private/qwaylandabstractdecoration_p.h>
 
 #include <QtGui/QGuiApplication>
@@ -31,6 +33,16 @@ QWaylandXdgSurface::Toplevel::Toplevel(QWaylandXdgSurface *xdgSurface)
     }
     requestWindowStates(window->windowStates());
     requestWindowFlags(window->flags());
+    if (auto transientParent = xdgSurface->window()->transientParent()) {
+        if (auto parentSurface =
+                    qobject_cast<QWaylandXdgSurface *>(transientParent->shellSurface())) {
+            set_parent(parentSurface->m_toplevel->object());
+            if (window->modality() != Qt::NonModal && m_xdgSurface->m_shell->m_xdgDialogWm) {
+                m_xdgDialog.reset(m_xdgSurface->m_shell->m_xdgDialogWm->getDialog(object()));
+                m_xdgDialog->set_modal();
+            }
+        }
+    }
 }
 
 QWaylandXdgSurface::Toplevel::~Toplevel()
@@ -157,7 +169,7 @@ void QWaylandXdgSurface::Toplevel::xdg_toplevel_configure(int32_t width, int32_t
 
 void QWaylandXdgSurface::Toplevel::xdg_toplevel_close()
 {
-    m_xdgSurface->m_window->window()->close();
+    QWindowSystemInterface::handleCloseEvent(m_xdgSurface->m_window->window());
 }
 
 void QWaylandXdgSurface::Toplevel::requestWindowFlags(Qt::WindowFlags flags)
@@ -177,6 +189,15 @@ void QWaylandXdgSurface::Toplevel::requestWindowStates(Qt::WindowStates states)
     // Re-send what's different from the applied state
     Qt::WindowStates changedStates = m_applied.states ^ states;
 
+    // Minimized state is not reported by the protocol, so always send it
+    if (states & Qt::WindowMinimized) {
+        set_minimized();
+        m_xdgSurface->window()->handleWindowStatesChanged(states & ~Qt::WindowMinimized);
+        // The internal window state whilst minimized is not maximised or fullscreen, but we don't want to
+        // update the compositors cached version of this state
+        return;
+    }
+
     if (changedStates & Qt::WindowMaximized) {
         if (states & Qt::WindowMaximized)
             set_maximized();
@@ -194,11 +215,7 @@ void QWaylandXdgSurface::Toplevel::requestWindowStates(Qt::WindowStates states)
             unset_fullscreen();
     }
 
-    // Minimized state is not reported by the protocol, so always send it
-    if (states & Qt::WindowMinimized) {
-        set_minimized();
-        m_xdgSurface->window()->handleWindowStatesChanged(states & ~Qt::WindowMinimized);
-    }
+
 }
 
 QtWayland::xdg_toplevel::resize_edge QWaylandXdgSurface::Toplevel::convertToResizeEdges(Qt::Edges edges)
@@ -227,9 +244,6 @@ QWaylandXdgSurface::Popup::~Popup()
         destroy();
 
     if (m_grabbing) {
-        auto *shell = m_xdgSurface->m_shell;
-        Q_ASSERT(shell->m_topmostGrabbingPopup == this);
-        shell->m_topmostGrabbingPopup = m_parentXdgSurface ? m_parentXdgSurface->m_popup : nullptr;
         m_grabbing = false;
 
         // Synthesize Qt enter/leave events for popup
@@ -238,8 +252,10 @@ QWaylandXdgSurface::Popup::~Popup()
             leave = m_xdgSurface->window()->window();
         QWindowSystemInterface::handleLeaveEvent(leave);
 
-        if (QWindow *enter = QGuiApplication::topLevelAt(QCursor::pos()))
-            QWindowSystemInterface::handleEnterEvent(enter, enter->mapFromGlobal(QCursor::pos()), QCursor::pos());
+        if (QWindow *enter = QGuiApplication::topLevelAt(QCursor::pos())) {
+            const auto pos = m_xdgSurface->window()->display()->waylandCursor()->pos();
+            QWindowSystemInterface::handleEnterEvent(enter, enter->handle()->mapFromGlobal(pos), pos);
+        }
     }
 }
 
@@ -261,7 +277,6 @@ void QWaylandXdgSurface::Popup::resetConfiguration()
 
 void QWaylandXdgSurface::Popup::grab(QWaylandInputDevice *seat, uint serial)
 {
-    m_xdgSurface->m_shell->m_topmostGrabbingPopup = this;
     xdg_popup::grab(seat->wl_seat(), serial);
     m_grabbing = true;
 }
@@ -273,7 +288,7 @@ void QWaylandXdgSurface::Popup::xdg_popup_configure(int32_t x, int32_t y, int32_
 
 void QWaylandXdgSurface::Popup::xdg_popup_popup_done()
 {
-    m_xdgSurface->m_window->window()->close();
+    QWindowSystemInterface::handleCloseEvent(m_xdgSurface->m_window->window());
 }
 
 QWaylandXdgSurface::QWaylandXdgSurface(QWaylandXdgShell *shell, ::xdg_surface *surface, QWaylandWindow *window)
@@ -292,11 +307,6 @@ QWaylandXdgSurface::QWaylandXdgSurface(QWaylandXdgShell *shell, ::xdg_surface *s
         setGrabPopup(transientParent, display->lastInputDevice(), display->lastInputSerial());
     } else {
         setToplevel();
-        if (transientParent) {
-            auto parentXdgSurface = qobject_cast<QWaylandXdgSurface *>(transientParent->shellSurface());
-            if (parentXdgSurface)
-                m_toplevel->set_parent(parentXdgSurface->m_toplevel->object());
-        }
     }
     setSizeHints();
 }
@@ -407,25 +417,30 @@ void QWaylandXdgSurface::propagateSizeHints()
 
 void QWaylandXdgSurface::setWindowGeometry(const QRect &rect)
 {
-    set_window_geometry(rect.x(), rect.y(), rect.width(), rect.height());
+    if (window()->isExposed())
+        set_window_geometry(rect.x(), rect.y(), rect.width(), rect.height());
 }
 
 void QWaylandXdgSurface::setSizeHints()
 {
     if (m_toplevel && m_window) {
-        const int minWidth = qMax(0, m_window->windowMinimumSize().width());
-        const int minHeight = qMax(0, m_window->windowMinimumSize().height());
-        int maxWidth = qMax(0, m_window->windowMaximumSize().width());
-        int maxHeight = qMax(0, m_window->windowMaximumSize().height());
-        if (maxWidth == QWINDOWSIZE_MAX)
-            maxWidth = 0;
-        if (maxHeight == QWINDOWSIZE_MAX)
-            maxHeight = 0;
+        const QMargins margins = m_window->windowContentMargins() - m_window->clientSideMargins();
+        const QSize minSize = m_window->windowMinimumSize().shrunkBy(margins);
+        const QSize maxSize = m_window->windowMaximumSize().shrunkBy(margins);
+        const int minWidth = qMax(0, minSize.width());
+        const int minHeight = qMax(0, minSize.height());
+        int maxWidth = qMax(0, maxSize.width());
+        int maxHeight = qMax(0, maxSize.height());
 
         // It will not change min/max sizes if invalid.
         if (minWidth > maxWidth || minHeight > maxHeight)
             return;
 
+        if (maxWidth == QWINDOWSIZE_MAX)
+            maxWidth = 0;
+        if (maxHeight == QWINDOWSIZE_MAX)
+            maxHeight = 0;
+
         m_toplevel->set_min_size(minWidth, minHeight);
         m_toplevel->set_max_size(maxWidth, maxHeight);
     }
@@ -574,20 +589,6 @@ void QWaylandXdgSurface::setPopup(QWaylandWindow *parent)
 
 void QWaylandXdgSurface::setGrabPopup(QWaylandWindow *parent, QWaylandInputDevice *device, int serial)
 {
-    auto parentXdgSurface = qobject_cast<QWaylandXdgSurface *>(parent->shellSurface());
-    auto *top = m_shell->m_topmostGrabbingPopup;
-
-    if (top && top->m_xdgSurface != parentXdgSurface) {
-        qCWarning(lcQpaWayland) << "setGrabPopup called with a parent," << parentXdgSurface
-                                << "which does not match the current topmost grabbing popup,"
-                                << top->m_xdgSurface << "According to the xdg-shell protocol, this"
-                                << "is not allowed. The wayland QPA plugin is currently handling"
-                                << "it by setting the parent to the topmost grabbing popup."
-                                << "Note, however, that this may cause positioning errors and"
-                                << "popups closing unxpectedly because xdg-shell mandate that child"
-                                << "popups close before parents";
-        parent = top->m_xdgSurface->m_window;
-    }
     setPopup(parent);
     m_popup->grab(device, serial);
 
@@ -605,8 +606,10 @@ void QWaylandXdgSurface::setGrabPopup(QWaylandWindow *parent, QWaylandInputDevic
     if (m_popup && m_popup->m_xdgSurface && m_popup->m_xdgSurface->window())
         enter = m_popup->m_xdgSurface->window()->window();
 
-    if (enter)
-        QWindowSystemInterface::handleEnterEvent(enter, enter->mapFromGlobal(QCursor::pos()), QCursor::pos());
+    if (enter) {
+        const auto pos = m_popup->m_xdgSurface->window()->display()->waylandCursor()->pos();
+        QWindowSystemInterface::handleEnterEvent(enter, enter->handle()->mapFromGlobal(pos), pos);
+    }
 }
 
 void QWaylandXdgSurface::xdg_surface_configure(uint32_t serial)
@@ -616,7 +619,7 @@ void QWaylandXdgSurface::xdg_surface_configure(uint32_t serial)
         // We have to do the initial applyConfigure() immediately, since that is the expose.
         applyConfigure();
         if (isExposed())
-            m_window->handleExpose(QRect(QPoint(), m_window->geometry().size()));
+            m_window->sendRecursiveExposeEvent();
     } else {
         // Later configures are probably resizes, so we have to queue them up for a time when we
         // are not painting to the window.
@@ -641,17 +644,20 @@ bool QWaylandXdgSurface::requestActivate()
             // focus stealing prevention indication, so requestXdgActivationToken call
             // is still necessary in that case.
             const auto wlWindow = focusWindow ? static_cast<QWaylandWindow*>(focusWindow->handle()) : m_window;
-            if (const auto xdgSurface = qobject_cast<QWaylandXdgSurface *>(wlWindow->shellSurface())) {
-                if (const auto seat = wlWindow->display()->lastInputDevice()) {
-                    const auto tokenProvider = activation->requestXdgActivationToken(
-                            wlWindow->display(), wlWindow->wlSurface(), seat->serial(), xdgSurface->m_appId);
-                    connect(tokenProvider, &QWaylandXdgActivationTokenV1::done, this,
-                            [this, tokenProvider](const QString &token) {
-                                m_shell->activation()->activate(token, window()->wlSurface());
-                                tokenProvider->deleteLater();
-                            });
-                    return true;
-                }
+
+            QString appId;
+            if (const auto xdgSurface = qobject_cast<QWaylandXdgSurface *>(wlWindow->shellSurface()))
+                appId = xdgSurface->m_appId;
+
+            if (const auto seat = wlWindow->display()->lastInputDevice()) {
+                const auto tokenProvider = activation->requestXdgActivationToken(
+                        wlWindow->display(), wlWindow->wlSurface(), seat->serial(), appId);
+                connect(tokenProvider, &QWaylandXdgActivationTokenV1::done, this,
+                        [this, tokenProvider](const QString &token) {
+                            m_shell->activation()->activate(token, window()->wlSurface());
+                            tokenProvider->deleteLater();
+                        });
+                return true;
             }
         }
     }
@@ -711,9 +717,18 @@ QString QWaylandXdgSurface::externWindowHandle()
         return QString();
     }
     if (!m_toplevel->m_exported) {
-        m_toplevel->m_exported.reset(m_shell->exporter()->exportToplevel(m_window->wlSurface()));
+        auto *exporterWrapper = static_cast<zxdg_exporter_v2 *>(
+                wl_proxy_create_wrapper(m_shell->exporter()->object()));
+        auto exportQueue = wl_display_create_queue(m_shell->display()->wl_display());
+        wl_proxy_set_queue(reinterpret_cast<wl_proxy *>(exporterWrapper), exportQueue);
+        m_toplevel->m_exported.reset(new QWaylandXdgExportedV2(
+                zxdg_exporter_v2_export_toplevel(exporterWrapper, m_window->wlSurface())));
         // handle events is sent immediately
-        m_shell->display()->forceRoundTrip();
+        wl_display_roundtrip_queue(m_shell->display()->wl_display(), exportQueue);
+
+        wl_proxy_set_queue(reinterpret_cast<wl_proxy *>(m_toplevel->m_exported->object()), nullptr);
+        wl_proxy_wrapper_destroy(exporterWrapper);
+        wl_event_queue_destroy(exportQueue);
     }
     return m_toplevel->m_exported->handle();
 }
@@ -743,6 +758,10 @@ void QWaylandXdgShell::handleRegistryGlobal(void *data, wl_registry *registry, u
     if (interface == QLatin1String(QWaylandXdgExporterV2::interface()->name)) {
         xdgShell->m_xdgExporter.reset(new QWaylandXdgExporterV2(registry, id, version));
     }
+
+    if (interface == QLatin1String(QWaylandXdgDialogWmV1::interface()->name)) {
+        xdgShell->m_xdgDialogWm.reset(new QWaylandXdgDialogWmV1(registry, id, version));
+    }
 }
 
 }
